/*******************************************************************************
  Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
 
  (c) Copyright 1996 - 2003 Gary Henderson (gary.henderson@ntlworld.com) and
                            Jerremy Koot (jkoot@snes9x.com)

  (c) Copyright 2002 - 2003 Matthew Kendora and
                            Brad Jorsch (anomie@users.sourceforge.net)
 

                      
  C4 x86 assembler and some C emulation code
  (c) Copyright 2000 - 2003 zsKnight (zsknight@zsnes.com),
                            _Demo_ (_demo_@zsnes.com), and
                            Nach (n-a-c-h@users.sourceforge.net)
                                          
  C4 C++ code
  (c) Copyright 2003 Brad Jorsch

  DSP-1 emulator code
  (c) Copyright 1998 - 2003 Ivar (ivar@snes9x.com), _Demo_, Gary Henderson,
                            John Weidman (jweidman@slip.net),
                            neviksti (neviksti@hotmail.com), and
                            Kris Bleakley (stinkfish@bigpond.com)
 
  DSP-2 emulator code
  (c) Copyright 2003 Kris Bleakley, John Weidman, neviksti, Matthew Kendora, and
                     Lord Nightmare (lord_nightmare@users.sourceforge.net

  OBC1 emulator code
  (c) Copyright 2001 - 2003 zsKnight, pagefault (pagefault@zsnes.com)
  Ported from x86 assembler to C by sanmaiwashi

  SPC7110 and RTC C++ emulator code
  (c) Copyright 2002 Matthew Kendora with research by
                     zsKnight, John Weidman, and Dark Force

  S-RTC C emulator code
  (c) Copyright 2001 John Weidman
  
  Super FX x86 assembler emulator code 
  (c) Copyright 1998 - 2003 zsKnight, _Demo_, and pagefault 

  Super FX C emulator code 
  (c) Copyright 1997 - 1999 Ivar and Gary Henderson.



 
  Specific ports contains the works of other authors. See headers in
  individual files.
 
  Snes9x homepage: http://www.snes9x.com
 
  Permission to use, copy, modify and distribute Snes9x in both binary and
  source form, for non-commercial purposes, is hereby granted without fee,
  providing that this license information and copyright notice appear with
  all copies and any derived work.
 
  This software is provided 'as-is', without any express or implied
  warranty. In no event shall the authors be held liable for any damages
  arising from the use of this software.
 
  Snes9x is freeware for PERSONAL USE only. Commercial users should
  seek permission of the copyright holders first. Commercial use includes
  charging money for Snes9x or software derived from Snes9x.
 
  The copyright holders request that bug fixes and improvements to the code
  should be forwarded to them so everyone can benefit from the modifications
  in future versions.
 
  Super NES and Super Nintendo Entertainment System are trademarks of
  Nintendo Co., Limited and its subsidiary companies.
*******************************************************************************/

/*******************************************************************************
  Snes9x custom (Snes9x for Mac OS) by zones
  
  This build was started from John Stiles' port.
*******************************************************************************/

#include "memmap.h"

#include "os9-prefix.h"
#include "os9-os.h"
#include "os9-snes9x.h"
#include "os9-stringtools.h"
#include "os9-cart.h"

static pascal void NavOpenCartEventHandler(const NavEventCallbackMessage, NavCBRecPtr, NavCallBackUserData);
static pascal void NavFreezeToEventHandler(const NavEventCallbackMessage, NavCBRecPtr, NavCallBackUserData);
static pascal void NavDefrostFromEventHandler(const NavEventCallbackMessage, NavCBRecPtr, NavCallBackUserData);
static pascal Boolean NavOpenCartFilter(AEDesc *, void *, NavCallBackUserData, NavFilterModes);
static pascal Boolean NavDefrostFromFilter(AEDesc *, void *, NavCallBackUserData, NavFilterModes);
static pascal Boolean NavDefrostFromPreview(NavCBRecPtr, NavCallBackUserData);

typedef struct
{
	NavDialogRef	nref;
	FSRef			ref;
	Boolean			reply;
	char			name[PATH_MAX + 1];
}	NavState;

bool8 PickCart(FSRef *ref)
{
	return (NavServicesOpen(ref));
}

bool8 NavServicesOpen(FSRef *ref)
{
	OSStatus					err;
	NavDialogCreationOptions	dialogOptions;
	NavEventUPP					eventUPP;
	NavObjectFilterUPP			filterUPP;
	NavState					nav;
	
	err = NavGetDefaultDialogCreationOptions(&dialogOptions);
	dialogOptions.optionFlags &= ~kNavAllowPreviews & ~kNavAllowMultipleFiles & ~kNavAllowStationery;
	dialogOptions.preferenceKey = 3;
	dialogOptions.clientName = CFSTR("SNES9X");
	dialogOptions.windowTitle = CFCopyLocalizedString(CFSTR("OpenROMMes"), "Open");
	dialogOptions.modality = kWindowModalityAppModal;
	dialogOptions.parentWindow = nil;
	
	eventUPP = NewNavEventUPP(NavOpenCartEventHandler);
	filterUPP = NewNavObjectFilterUPP(NavOpenCartFilter);
	
	err = NavCreateChooseFileDialog(&dialogOptions, nil, eventUPP, nil, filterUPP, &nav, &(nav.nref));
	if (err == noErr)
	{
		err = NavDialogRun(nav.nref);
		if (err)
			NavDialogDispose(nav.nref);
	}
	
	DisposeNavObjectFilterUPP(filterUPP);
	DisposeNavEventUPP(eventUPP);
	
	CFRelease(dialogOptions.windowTitle);
	
	if (err)
		return false;
	else
	{
		if (nav.reply)
		{
			*ref = nav.ref;
			return true;
		}
		else
			return false;
	}
}

static pascal void NavOpenCartEventHandler(const NavEventCallbackMessage callBackSelector, NavCBRecPtr callBackParms, NavCallBackUserData callBackUD)
{
	OSStatus			err;
	NavReplyRecord		reply;
	NavUserAction		userAction;
	NavState			*nav;
	DialogRef			dialog;
	ControlRef 			control;
	ControlFontStyleRec fontStyle;
	AEDesc				resultDesc;
	Handle   			rditl = nil;
	SInt32 				count;
	UInt16				firstControl;
	
	nav = (NavState *) callBackUD;
	
	switch (callBackSelector)
	{
		case kNavCBEvent:
			break;
		
		case kNavCBStart:			
			nav->reply = false;
			dialog = GetDialogFromWindow(callBackParms->window);
	
			rditl = GetResource('DITL', 950);
			err = NavCustomControl(callBackParms->context, kNavCtlAddControlList, rditl);
			err = NavCustomControl(callBackParms->context, kNavCtlGetFirstControlID, &firstControl);

			fontStyle.flags = kControlUseFontMask;
			fontStyle.font  = kControlFontSmallSystemFont;
			
			err = GetDialogItemAsControl(dialog, firstControl + 1, &control);
			err = SetControlFontStyle(control, &fontStyle);			
			switch (romDetect)
			{
				case kAutoROMType:			SetControlValue(control, 1); break;
				case kLoROMForce:  			SetControlValue(control, 3); break;
				case kHiROMForce:  			SetControlValue(control, 4);
			}

			err = GetDialogItemAsControl(dialog, firstControl + 2, &control);
			err = SetControlFontStyle(control, &fontStyle);
			switch (interleaveDetect)
			{
				case kAutoInterleave:		SetControlValue(control, 1); break;
				case kNoInterleaveForce:	SetControlValue(control, 3); break;
				case kInterleaveForce:		SetControlValue(control, 4); break;
				case kInterleave2Force:		SetControlValue(control, 5);
			}
			
			err = GetDialogItemAsControl(dialog, firstControl + 3, &control);
			err = SetControlFontStyle(control, &fontStyle);	
			switch (videoDetect)
			{
				case kAutoVideo:			SetControlValue(control, 1); break;
				case kPALForce:				SetControlValue(control, 3); break;
				case kNTSCForce:			SetControlValue(control, 4);
			}
				
			err = GetDialogItemAsControl(dialog, firstControl + 4, &control);
			err = SetControlFontStyle(control, &fontStyle);			
			switch (headerDetect)
			{
				case kAutoHeader:			SetControlValue(control, 1); break;
				case kNoHeaderForce:		SetControlValue(control, 3); break;
				case kHeaderForce:			SetControlValue(control, 4);
			}

			fontStyle.flags = kControlUseFontMask | kControlUseJustMask;
			fontStyle.font  = kControlFontSmallSystemFont;
			fontStyle.just  = teFlushRight;

			for (int i = 5; i <= 8; i++)
			{
				err = GetDialogItemAsControl(dialog, firstControl + i, &control);
				err = SetControlFontStyle(control, &fontStyle);			
			}
			
			break;
			
		case kNavCBAccept:
			dialog = GetDialogFromWindow(callBackParms->window);
			err = NavCustomControl(callBackParms->context, kNavCtlGetFirstControlID, &firstControl);
			
			err = GetDialogItemAsControl(dialog, firstControl + 1, &control);
			switch (GetControlValue(control))
			{
				case 1: romDetect        = kAutoROMType; 		break;
				case 3: romDetect        = kLoROMForce;  		break;
				case 4: romDetect        = kHiROMForce;
			}

			err = GetDialogItemAsControl(dialog, firstControl + 2, &control);
			switch (GetControlValue(control))
			{
				case 1: interleaveDetect = kAutoInterleave;		break;
				case 3: interleaveDetect = kNoInterleaveForce;	break;
				case 4: interleaveDetect = kInterleaveForce;	break;
				case 5: interleaveDetect = kInterleave2Force;
			}
			
			err = GetDialogItemAsControl(dialog, firstControl + 3, &control);
			switch (GetControlValue(control))
			{
				case 1: videoDetect      = kAutoVideo; 			break;
				case 3: videoDetect      = kPALForce;  			break;
				case 4: videoDetect      = kNTSCForce;
			}

			err = GetDialogItemAsControl(dialog, firstControl + 4, &control);
			switch (GetControlValue(control))
			{
				case 1: headerDetect     = kAutoHeader; 		break;
				case 3: headerDetect     = kNoHeaderForce;  	break;
				case 4: headerDetect     = kHeaderForce;
			}

			break;
	
		case kNavCBCustomize:
			if ((callBackParms->customRect.right == 0) && (callBackParms->customRect.bottom == 0))
			{
				callBackParms->customRect.bottom = callBackParms->customRect.top  + 60;
				callBackParms->customRect.right  = callBackParms->customRect.left + 425;
			}

			break;
			
		case kNavCBUserAction:
			userAction = NavDialogGetUserAction(callBackParms->context);
			switch (userAction)
 			{
  				case kNavUserActionChoose:      
					err = NavDialogGetReply(callBackParms->context, &reply);
					if (err == noErr)
					{
						err = AECountItems(&(reply.selection), &count);
						if ((err == noErr) && (count == 1))
						{
							err = AEGetNthDesc(&(reply.selection), 1, typeFSS, nil, &resultDesc);
							if (err == noErr)
							{
								FSSpec	spec;
								
								err = AEGetDescData(&resultDesc, &spec, sizeof(FSSpec));
								if (err == noErr)
								{
									err = FSpMakeFSRef(&spec, &(nav->ref));
									if (err == noErr)
										nav->reply = true;
								}

								err = AEDisposeDesc(&resultDesc);
							}
				        }
	  					
	  					err = NavDisposeReply(&reply);
	  				}  
            }
            
            break;
			
		case kNavCBTerminate:
			ReleaseResource(rditl);
			NavDialogDispose(nav->nref);
	}		
}

static pascal Boolean NavOpenCartFilter(AEDesc *theItem, void *ninfo, NavCallBackUserData callbackUD, NavFilterModes filterMode)
{
	#pragma unused (ninfo, callbackUD, filterMode)
	
	OSStatus	err;
	AEDesc 		resultDesc;
	Boolean		result = true;

	err = AECoerceDesc(theItem, typeFSS, &resultDesc);
	if (err == noErr)
	{
		FSRef	ref;
		FSSpec	spec;

		err = AEGetDescData(&resultDesc, &spec, sizeof(FSSpec));
		if (err == noErr)
		{
			err = FSpMakeFSRef(&spec, &ref);
			if (err == noErr)
			{
				FSCatalogInfo	catinfo;
				HFSUniStr255	unistr;
				
				err = FSGetCatalogInfo(&ref, kFSCatInfoNodeFlags, &catinfo, &unistr, nil, nil);
				if ((err == noErr) && !(catinfo.nodeFlags & kFSNodeIsDirectoryMask) && (unistr.length > 4))
				{
					UInt16	i = unistr.length;
					
					if  ((unistr.unicode[i - 4] == '.') &&
					   (((unistr.unicode[i - 3] == 's') && (unistr.unicode[i - 2] == 'r') && (unistr.unicode[i - 1] == 'm')) ||
						((unistr.unicode[i - 3] == 'f') && (unistr.unicode[i - 2] == 'r') && (unistr.unicode[i - 1] == 'z')) ||
						((unistr.unicode[i - 3] == 's') && (unistr.unicode[i - 2] == 'p') && (unistr.unicode[i - 1] == 'c')) ||
						((unistr.unicode[i - 3] == 'r') && (unistr.unicode[i - 2] == 't') && (unistr.unicode[i - 1] == 'c')) ||
						((unistr.unicode[i - 3] == 'd') && (unistr.unicode[i - 2] == 'a') && (unistr.unicode[i - 1] == 't')) ||
						((unistr.unicode[i - 3] == 'c') && (unistr.unicode[i - 2] == 'h') && (unistr.unicode[i - 1] == 't')) ||
						((unistr.unicode[i - 3] == 'i') && (unistr.unicode[i - 2] == 'p') && (unistr.unicode[i - 1] == 's')) ||
						((unistr.unicode[i - 3] == 'p') && (unistr.unicode[i - 2] == 'n') && (unistr.unicode[i - 1] == 'g'))))
						result = false;
				}
			}
		}

		AEDisposeDesc(&resultDesc);
	}
	
	return result;
}

bool8 NavFreezeTo(char *path)
{
	OSStatus					err;
	NavDialogCreationOptions	dialogOptions;
	NavEventUPP					eventUPP;
	NavState					nav;
	CFStringRef					numRef, romRef, baseRef;
	CFMutableStringRef			mesRef, saveRef;
	SInt32						replaceAt;
	char						drive[_MAX_DRIVE], dir[_MAX_DIR], fname[_MAX_FNAME], ext[_MAX_EXT];
	
	_splitpath(Memory.ROMFilename, drive, dir, fname, ext);
	romRef  = CFStringCreateWithCString(kCFAllocatorDefault, fname, MAC_PATH_ENCODING);
	numRef  = CFCopyLocalizedString(CFSTR("FreezeToPos"), "1");
	baseRef = CFCopyLocalizedString(CFSTR("FreezeToMes"), "Freeze");
	mesRef  = CFStringCreateMutableCopy(kCFAllocatorDefault, 0, baseRef);
	saveRef = CFStringCreateMutableCopy(kCFAllocatorDefault, 0, romRef);
	replaceAt = CFStringGetIntValue(numRef);
	CFStringReplace(mesRef, CFRangeMake(replaceAt - 1, 1), romRef);
	CFStringAppendCString(saveRef, ".frz", CFStringGetSystemEncoding());
	
	err = NavGetDefaultDialogCreationOptions(&dialogOptions);
	dialogOptions.preferenceKey = 4;
	dialogOptions.clientName = CFSTR("SNES9X");
	dialogOptions.windowTitle = mesRef;
	dialogOptions.saveFileName = saveRef;
	dialogOptions.modality = kWindowModalityAppModal;
	dialogOptions.parentWindow = nil;
	
	eventUPP = NewNavEventUPP(NavFreezeToEventHandler);
	
	err = NavCreatePutFileDialog(&dialogOptions, 'SAVE', '~9X~', eventUPP, &nav, &(nav.nref));
	if (err == noErr)
	{
		err = NavDialogRun(nav.nref);
        if (err)
        	NavDialogDispose(nav.nref);
	}
	
	DisposeNavEventUPP(eventUPP);
	
	CFRelease(saveRef);
	CFRelease(mesRef);
	CFRelease(baseRef);
	CFRelease(numRef);
	CFRelease(romRef);
	
	if (err)
		return false;
	else
	{
		if (nav.reply)
		{
			err = FSRefMakePath(&(nav.ref), (unsigned char *) path, PATH_MAX);
			strcat(path, MAC_PATH_SEPARATOR);
			strcat(path, nav.name);
			
			return true;
		}
		else
			return false;
	}
}

static pascal void NavFreezeToEventHandler(const NavEventCallbackMessage callBackSelector, NavCBRecPtr callBackParms, NavCallBackUserData callBackUD)
{
	OSStatus			err;
	NavReplyRecord		reply;
	NavUserAction		userAction;
	NavState			*nav;
	AEDesc				resultDesc;
	
	nav = (NavState *) callBackUD;
	
	switch (callBackSelector)
	{
		case kNavCBEvent:
			break;
			
		case kNavCBStart:
			nav->reply = false;
			
			break;
		
		case kNavCBUserAction:
			userAction = NavDialogGetUserAction(callBackParms->context);
			switch (userAction)
			{
				case kNavUserActionSaveAs:
					err = NavDialogGetReply(callBackParms->context, &reply);
					if (err == noErr)
					{
						err = AEGetNthDesc(&(reply.selection), 1, typeFSS, nil, &resultDesc);
						if (err == noErr)
						{
							FSSpec	spec;
							
							err = AEGetDescData(&resultDesc, &spec, sizeof(FSSpec));
							if (err == noErr)
							{
								err = FSpMakeFSRef(&spec, &(nav->ref));
								if (err == noErr)
								{
									Str255	pstr;
									
									err = NavCustomControl(callBackParms->context, kNavCtlGetEditFileName, pstr);
									if (err == noErr)
									{
										ConvertPString(pstr, nav->name);
										nav->reply = true;
									}
								}
							}

							err = AEDisposeDesc(&resultDesc);
				        }
						
						err = NavDisposeReply(&reply);
					}
            }
 
            break;
			
		case kNavCBTerminate:
			NavDialogDispose(nav->nref);
	}		
}

bool8 NavDefrostFrom(char *path)
{
	OSStatus					err;
	NavDialogCreationOptions	dialogOptions;
	NavEventUPP					eventUPP;
	NavObjectFilterUPP			filterUPP;
	NavPreviewUPP				previewUPP;
	NavState					nav;
	CFStringRef					numRef, romRef, baseRef;
	CFMutableStringRef			mesRef;
	SInt32						replaceAt;
	char						drive[_MAX_DRIVE], dir[_MAX_DIR], fname[_MAX_FNAME], ext[_MAX_EXT];

	_splitpath(Memory.ROMFilename, drive, dir, fname, ext);
	romRef  = CFStringCreateWithCString(kCFAllocatorDefault, fname, MAC_PATH_ENCODING);
	numRef  = CFCopyLocalizedString(CFSTR("DefrostFromPos"), "1");
	baseRef = CFCopyLocalizedString(CFSTR("DefrostFromMes"), "Defrost");
	mesRef  = CFStringCreateMutableCopy(kCFAllocatorDefault, 0, baseRef);
	replaceAt = CFStringGetIntValue(numRef);
	CFStringReplace(mesRef, CFRangeMake(replaceAt - 1, 1), romRef);
	
	err = NavGetDefaultDialogCreationOptions(&dialogOptions);
	dialogOptions.optionFlags &= ~kNavAllowMultipleFiles & ~kNavAllowStationery;
	dialogOptions.preferenceKey = 5;
	dialogOptions.clientName = CFSTR("SNES9X");
	dialogOptions.windowTitle = mesRef;
	dialogOptions.modality = kWindowModalityAppModal;
	dialogOptions.parentWindow = nil;

	eventUPP = NewNavEventUPP(NavDefrostFromEventHandler);
	filterUPP = NewNavObjectFilterUPP(NavDefrostFromFilter);
	previewUPP = NewNavPreviewUPP(NavDefrostFromPreview);
	
	err = NavCreateChooseFileDialog(&dialogOptions, nil, eventUPP, previewUPP, filterUPP, &nav, &(nav.nref));
	if (err == noErr)
	{
		err = NavDialogRun(nav.nref);
        if (err)
        	NavDialogDispose(nav.nref);
	}
	
	DisposeNavPreviewUPP(previewUPP);
	DisposeNavObjectFilterUPP(filterUPP);
	DisposeNavEventUPP(eventUPP);
	
	CFRelease(mesRef);
	CFRelease(baseRef);
	CFRelease(numRef);
	CFRelease(romRef);
	
	if (err)
		return false;
	else
	{
		if (nav.reply)
		{
			err = FSRefMakePath(&(nav.ref), (unsigned char *) path, PATH_MAX);
			return true;
		}
		else
			return false;
	}
}

static pascal void NavDefrostFromEventHandler(const NavEventCallbackMessage callBackSelector, NavCBRecPtr callBackParms, NavCallBackUserData callBackUD)
{
	OSStatus			err;
	NavReplyRecord		reply;
	NavUserAction		userAction;
	NavState			*nav;
	AEDesc				resultDesc;
	SInt32				count;
	
	nav = (NavState *) callBackUD;
	
	switch (callBackSelector)
	{
		case kNavCBEvent:
			break;
			
		case kNavCBStart:
			nav->reply = false;
			
			break;
		
		case kNavCBUserAction:
			userAction = NavDialogGetUserAction(callBackParms->context);
			switch (userAction)
			{
				case kNavUserActionChoose:
					err = NavDialogGetReply(callBackParms->context, &reply);
					if (err == noErr)
					{
						err = AECountItems(&(reply.selection), &count);
						if ((err == noErr) && (count == 1))
						{
							err = AEGetNthDesc(&(reply.selection), 1, typeFSS, nil, &resultDesc);
							if (err == noErr)
							{
								FSSpec	spec;
								
								err = AEGetDescData(&resultDesc, &spec, sizeof(FSSpec));
								if (err == noErr)
								{
									err = FSpMakeFSRef(&spec, &(nav->ref));
									if (err == noErr)
										nav->reply = true;
								}
	
								err = AEDisposeDesc(&resultDesc);
							}
						}
						
						err = NavDisposeReply(&reply);
					}
            }

            break;
			
		case kNavCBTerminate:
			NavDialogDispose(nav->nref);
	}		
}

static pascal Boolean NavDefrostFromFilter(AEDesc *theItem, void *ninfo, NavCallBackUserData callbackUD, NavFilterModes filterMode)
{
	#pragma unused (ninfo, callbackUD, filterMode)
	
	OSStatus	err;
	AEDesc 		resultDesc;
	Boolean		result = true;

	err = AECoerceDesc(theItem, typeFSS, &resultDesc);
	if (err == noErr)
	{
		FSRef	ref;
		FSSpec	spec;

		err = AEGetDescData(&resultDesc, &spec, sizeof(FSSpec));
		if (err == noErr)
		{
			err = FSpMakeFSRef(&spec, &ref);
			if (err == noErr)
			{
				FSCatalogInfo	catinfo;
				HFSUniStr255	unistr;			

				err = FSGetCatalogInfo(&ref, kFSCatInfoNodeFlags | kFSCatInfoFinderInfo, &catinfo, &unistr, nil, nil);
				if ((err == noErr) && !(catinfo.nodeFlags & kFSNodeIsDirectoryMask))
				{
					if (((FileInfo *) &catinfo.finderInfo)->fileType != 'SAVE')
					{
						UInt16	i = unistr.length;
						
						if (i < 4)
							result = false;
						else
						if ((unistr.unicode[i - 4] != '.') || (unistr.unicode[i - 3] != 'f') || (unistr.unicode[i - 2] != 'r') || (unistr.unicode[i - 1] != 'z'))
							result = false;
					}
				}
			}
		}
						
		AEDisposeDesc(&resultDesc);
	}
	
	return result;
}

static pascal Boolean NavDefrostFromPreview(NavCBRecPtr callBackParms, NavCallBackUserData callBackUD)
{
	#pragma unused (callBackUD)
	
	OSStatus	err;
	Boolean		previewShowing, result = false;
	
	err = NavCustomControl(callBackParms->context, kNavCtlIsPreviewShowing, &previewShowing);
	if ((err == noErr) && previewShowing)
	{
		AEDesc	resultDesc;
		
		err = AECoerceDesc((AEDesc *) callBackParms->eventData.eventDataParms.param, typeFSS, &resultDesc);
		if (err == noErr)
		{
			FSRef	ref;
			FSSpec	sc;
			
			err = AEGetDescData(&resultDesc, &sc, sizeof(FSSpec));
			if (err == noErr)
			{
				err = FSpMakeFSRef(&sc, &ref);
				if (err == noErr)
				{
					FSCatalogInfo	catinfo;
					FSSpec			spec;
					
					err = FSGetCatalogInfo(&ref, kFSCatInfoNodeFlags | kFSCatInfoContentMod, &catinfo, nil, &spec, nil);
					if ((err == noErr) && !(catinfo.nodeFlags & kFSNodeIsDirectoryMask))
					{
						LocalDateTime		localtime;
						LongDateTime		ldtime;
						CFMutableStringRef	sref;
						Str255				pstr;
						CGrafPtr			oldPort;
						PicHandle			myPict;
						RGBColor			foreColor;
						Rect				textbox, rct;
						short				refNum;

						sref = CFStringCreateMutable(kCFAllocatorDefault, 0);
						
						ConvertUTCToLocalDateTime(&(catinfo.contentModDate), &localtime);				
						ldtime = (LongDateTime) ((((UInt64) localtime.highSeconds) << 32) | ((UInt64) localtime.lowSeconds));
						LongDateString(&ldtime, shortDate, pstr, nil);			
						CFStringAppendPascalString(sref, pstr,   CFStringGetSystemEncoding());
						CFStringAppendPascalString(sref, "\p  ", CFStringGetSystemEncoding());
						LongTimeString(&ldtime, true, pstr, nil);
						CFStringAppendPascalString(sref, pstr,   CFStringGetSystemEncoding());
					
						GetPort(&oldPort);
						SetPortWindowPort(callBackParms->window);
						
						rct.left   = ((callBackParms->previewRect.right  - callBackParms->previewRect.left - 128) >> 1) + callBackParms->previewRect.left;
						rct.top    = ((callBackParms->previewRect.bottom - callBackParms->previewRect.top  - 132) >> 1) + callBackParms->previewRect.top;
						rct.right  = rct.left + 128;
						rct.bottom = rct.top  + 120;
						
						refNum = FSpOpenResFile(&spec, fsRdPerm);
						err = ResError();
						myPict = GetPicture(128);			
						if (myPict)
						{
							DrawPicture(myPict, &rct); 
							ReleaseResource((Handle) myPict);
						}
						if (err == noErr)
							CloseResFile(refNum);
							
						GetForeColor(&foreColor);
						ForeColor(blackColor);
						SetRect(&textbox, rct.left, rct.top + 123, rct.right, rct.top + 145);
						DrawThemeTextBox(sref, kThemeSmallSystemFont, kThemeStateActive, false, &textbox, teCenter, nil);
						RGBForeColor(&foreColor);
						
						SetPort(oldPort);
						
						CFRelease(sref);
						
						result = true;
					}
				}
			}
			
			AEDisposeDesc(&resultDesc);
		}
	}
	
	return result;
}
