/*******************************************************************************
  Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
 
  (c) Copyright 1996 - 2003 Gary Henderson (gary.henderson@ntlworld.com) and
                            Jerremy Koot (jkoot@snes9x.com)

  (c) Copyright 2002 - 2003 Matthew Kendora and
                            Brad Jorsch (anomie@users.sourceforge.net)
 

                      
  C4 x86 assembler and some C emulation code
  (c) Copyright 2000 - 2003 zsKnight (zsknight@zsnes.com),
                            _Demo_ (_demo_@zsnes.com), and
                            Nach (n-a-c-h@users.sourceforge.net)
                                          
  C4 C++ code
  (c) Copyright 2003 Brad Jorsch

  DSP-1 emulator code
  (c) Copyright 1998 - 2003 Ivar (ivar@snes9x.com), _Demo_, Gary Henderson,
                            John Weidman (jweidman@slip.net),
                            neviksti (neviksti@hotmail.com), and
                            Kris Bleakley (stinkfish@bigpond.com)
 
  DSP-2 emulator code
  (c) Copyright 2003 Kris Bleakley, John Weidman, neviksti, Matthew Kendora, and
                     Lord Nightmare (lord_nightmare@users.sourceforge.net

  OBC1 emulator code
  (c) Copyright 2001 - 2003 zsKnight, pagefault (pagefault@zsnes.com)
  Ported from x86 assembler to C by sanmaiwashi

  SPC7110 and RTC C++ emulator code
  (c) Copyright 2002 Matthew Kendora with research by
                     zsKnight, John Weidman, and Dark Force

  S-RTC C emulator code
  (c) Copyright 2001 John Weidman
  
  Super FX x86 assembler emulator code 
  (c) Copyright 1998 - 2003 zsKnight, _Demo_, and pagefault 

  Super FX C emulator code 
  (c) Copyright 1997 - 1999 Ivar and Gary Henderson.



 
  Specific ports contains the works of other authors. See headers in
  individual files.
 
  Snes9x homepage: http://www.snes9x.com
 
  Permission to use, copy, modify and distribute Snes9x in both binary and
  source form, for non-commercial purposes, is hereby granted without fee,
  providing that this license information and copyright notice appear with
  all copies and any derived work.
 
  This software is provided 'as-is', without any express or implied
  warranty. In no event shall the authors be held liable for any damages
  arising from the use of this software.
 
  Snes9x is freeware for PERSONAL USE only. Commercial users should
  seek permission of the copyright holders first. Commercial use includes
  charging money for Snes9x or software derived from Snes9x.
 
  The copyright holders request that bug fixes and improvements to the code
  should be forwarded to them so everyone can benefit from the modifications
  in future versions.
 
  Super NES and Super Nintendo Entertainment System are trademarks of
  Nintendo Co., Limited and its subsidiary companies.
*******************************************************************************/

/*******************************************************************************
  Snes9x custom (Snes9x for Mac OS) by zones
  
  This build was started from John Stiles' port.
*******************************************************************************/

#include "snes9x.h"
#include "display.h"
#include "memmap.h"

#include "os9-prefix.h"
#include "os9-dialog.h"
#include "os9-os.h"
#include "os9-stringtools.h"
#include "os9-file.h"

extern char	pfold[9];

extern "C" char * osd_GetPackDir(void);

static OSErr FindSNESFolder(FSRef *, char *, const char *);
static int S9xCompareSDD1IndexEntries(const void *, const void *);
static const char * S9xGetPackDirectory(void);

static char	gPackFolderPath[PATH_MAX + 1];

static OSErr FindSNESFolder(FSRef *folderRef, char *folderPath, const char *folderName)
{
	OSStatus	err;
	CFURLRef	burl, purl;
	CFStringRef	fstr;
	FSRef		pref;
	UniChar		buffer[PATH_MAX + 1];
	Boolean		r;
	
	fstr = CFStringCreateWithCString(kCFAllocatorDefault, folderName, CFStringGetSystemEncoding());
	CFStringGetCharacters(fstr, CFRangeMake(0, CFStringGetLength(fstr)), buffer);

	burl = CFBundleCopyBundleURL(CFBundleGetMainBundle());
	purl = CFURLCreateCopyDeletingLastPathComponent(kCFAllocatorDefault, burl);
	r    = CFURLGetFSRef(purl, &pref);
	
	err = FSMakeFSRefUnicode(&pref, CFStringGetLength(fstr), buffer, kTextEncodingUnicodeDefault, folderRef);
 	if (err == dirNFErr || err == fnfErr) 
		err = FSCreateDirectoryUnicode(&pref, CFStringGetLength(fstr), buffer, kFSCatInfoNone, nil, folderRef, nil, nil);
	
	if (err != noErr && !folderWarning)
	{
		AppearanceAlert(kAlertCautionAlert, kFolderFail, kFolderHint);
		folderWarning = true;
	}
	else
		err = FSRefMakePath(folderRef, (unsigned char *) folderPath, PATH_MAX);
		
	CFRelease(purl);
	CFRelease(burl);
	CFRelease(fstr);

	return err;
}

void ChangeTypeAndCreator(const char *path, OSType type, OSType creator)
{
	OSErr	err;
	FSRef	ref;

	err = ClassicFSPathMakeFSRef((unsigned char *) path, &ref, nil);
	if (!err)
	{
		FSCatalogInfo	catinfo;
		
		((FileInfo *) &catinfo.finderInfo)->finderFlags   = 0;
		((FileInfo *) &catinfo.finderInfo)->reservedField = 0;
		((FileInfo *) &catinfo.finderInfo)->location.v    = 0;
		((FileInfo *) &catinfo.finderInfo)->location.h    = 0;
		((FileInfo *) &catinfo.finderInfo)->fileCreator   = creator;
    	((FileInfo *) &catinfo.finderInfo)->fileType      = type;
		
		err = FSSetCatalogInfo(&ref, kFSCatInfoFinderInfo, &catinfo);
	}
}

Boolean IsLockedMedia(FSVolumeRefNum volume)
{
    OSErr			err;
	FSVolumeInfo	info;
	
	err = FSGetVolumeInfo(volume, 0, nil, kFSVolInfoFlags, &info, nil, nil);
    if (err == noErr)
	{
		if ((info.flags & kFSVolFlagHardwareLockedMask) || (info.flags & kFSVolFlagSoftwareLockedMask))
			return true;
	}
	
	return false;
}

const char * S9xGetFilename(const char *inExt)
{
	static char	filePath[PATH_MAX + 1];

	OSErr	err;
	FSRef	ref;
	char	folderName[16];
	char	drive[_MAX_DRIVE], dir[_MAX_DIR], fname[_MAX_FNAME], ext[_MAX_EXT];
	
	folderName[0] = filePath[0] = 0;
	
	if (strlen(inExt) < 4)
		return filePath;
	
	switch (*(long *) (inExt + strlen(inExt) - 4))
	{
		case '.srm':
		case '.dat':
		case '.rtc':
			strcpy(folderName, "SRAMs");
			break;
		
		case '.frz':
			strcpy(folderName, "Freezes");
			break;

		case '.spc':
			strcpy(folderName, "SPCs");
			break;
		
		case '.cht':
			strcpy(folderName, "Cheats");
			break;
			
		case '.ips':
			strcpy(folderName, "IPSs");
			break;
			
		case '.png':
			strcpy(folderName, "Screenshots");
			break;
			
		case '.pac':	// dummy
			strcpy(folderName, "Packs");
	}

	if (folderName[0] && ((saveInROMFolder != 1) || lockedROMMedia))
	{
		err = FindSNESFolder(&ref, filePath, folderName);
			
		if (err == noErr)
		{			
			_splitpath(Memory.ROMFilename, drive, dir, fname, ext);
			strcat(filePath, MAC_PATH_SEPARATOR);
			strcat(filePath, fname);
			strcat(filePath, inExt);
		}
		else
		{
			_splitpath(Memory.ROMFilename, drive, dir, fname, ext);
			_makepath(filePath, drive, dir, fname, inExt);
		}
	}
	else
	{
	    _splitpath(Memory.ROMFilename, drive, dir, fname, ext);
		_makepath(filePath, drive, dir, fname, inExt);
	}

	return filePath;
}

const char * S9xGetSPCFilename(void)
{
	char	spcExt[16];
	
	sprintf(spcExt, ".%03d.spc", (int) spcFileCount);
	
	spcFileCount++;
	if (spcFileCount == 999)
		spcFileCount = 0;

	return S9xGetFilename(spcExt);
}

const char * S9xGetPNGFilename(void)
{
	char	pngExt[16];
	
	sprintf(pngExt, ".%03d.png", (int) pngFileCount);
	
	pngFileCount++;
	if (pngFileCount == 999)
		pngFileCount = 0;

	return S9xGetFilename(pngExt);
}

const char * S9xGetFreezeFilename(int which)
{
	char	frzExt[16];
	
	sprintf(frzExt, ".%03d.frz", which);
	
	return S9xGetFilename(frzExt);
}

extern "C" const char * S9xGetFilenameInc(const char *inExt)
{
	if (strlen(inExt) < 4)
		return nil;
	
	switch (*(long *) (inExt + strlen(inExt) - 4))
	{
		case '.spc':
			return S9xGetSPCFilename();
		
		case '.png':
			return S9xGetPNGFilename();
	}
	
	return nil;
}

bool8 S9xOpenSnapshotFile(const char *fname, bool8 read_only, STREAM *file)
{
    if (read_only)
    {
		if (0 != (*file = OPEN_STREAM(fname, "rb")))
		    return true;
    }
    else
    {
		if (0 != (*file = OPEN_STREAM(fname, "wb")))
		    return true;
    }

    return false;
}

void S9xCloseSnapshotFile(STREAM file)
{
    CLOSE_STREAM(file);
}

const char * S9xBasename(const char *in)
{
	#pragma unused (in)

	return "";
}

static int S9xCompareSDD1IndexEntries(const void *p1, const void *p2)
{
    return (*(uint32 *) p1 - *(uint32 *) p2);
}

void S9xLoadSDD1Data(void)
{
	FILE	*fs;
	int		len;
	char	path[PATH_MAX + 1], index[PATH_MAX + 1], data[PATH_MAX + 1];
	
    Memory.FreeSDD1Data();
	
	if (macSDD1Pack == 1)
	{
		Settings.SDD1Pack = true;
		//printf("Using on-the fly S-DD1 decompression...\n");
		return;
	}
	
	Settings.SDD1Pack = false;

    strcpy(path, S9xGetPackDirectory());
	
    strcpy(index, path);
    strcat(index, "SDD1GFX.idx");
    strcpy(data,  path);
    strcat(data,  "SDD1GFX.dat");

	len = 0;
	
    fs = fopen(index, "rb");
    if (fs)
    {
        fseek(fs, 0, SEEK_END);
        len = ftell(fs);
        rewind(fs);
        Memory.SDD1Index = (uint8 *) malloc(len);
        fread(Memory.SDD1Index, 1, len, fs);
        fclose(fs);

        Memory.SDD1Entries = len / 12;
		
		fs = fopen(data, "rb");
        if (!fs)
        {
            free((char *) Memory.SDD1Index);
            Memory.SDD1Index = nil;
            Memory.SDD1Entries = 0;
			S9xMessage(0, 0, "Graphics Pack not found!");
			
			if (macSDD1Pack == 2)
			{
				//printf("Using on-the fly S-DD1 decompression...\n");
				Settings.SDD1Pack = true;
			}
		}
        else
        {
 			uint8			*ptr;
			unsigned int	i;
			
			fseek(fs, 0, SEEK_END);
            len = ftell(fs);
            rewind(fs);
            Memory.SDD1Data = (uint8 *) malloc(len);
            fread(Memory.SDD1Data, 1, len, fs);
            fclose(fs);
			
			ptr = Memory.SDD1Index;
			
          	for (i = 0; i < Memory.SDD1Entries; i++, ptr += 12)
		    {
				SWAP_DWORD((*(uint32 *) (ptr + 0)));
				SWAP_DWORD((*(uint32 *) (ptr + 4)));
				SWAP_DWORD((*(uint32 *) (ptr + 8)));
		    }

			qsort(Memory.SDD1Index, Memory.SDD1Entries, 12, S9xCompareSDD1IndexEntries);
        }
    }
	else
	{
		S9xMessage(0, 0, "Graphics Pack not found!");
		if (macSDD1Pack == 2)
		{
			//printf("Using on-the fly S-DD1 decompression...\n");
			Settings.SDD1Pack = true;
		}
	}
}

static const char * S9xGetPackDirectory(void)
{
	OSStatus	err;
	FSRef		ref;
	Boolean		isFolder, wasAliased;
	int			n;
	char		drive[_MAX_DRIVE], dir[_MAX_DIR], fname[_MAX_FNAME], ext[_MAX_EXT];

	static char	path[PATH_MAX + 1];
	static char	dummy[] = "~/";
	
	pfold[0] = 0;	// unused

	gPackFolderPath[0] = 0;

	strcpy(path, S9xGetFilename(".pac"));	
	n = strlen(path);
	if (n < 4)
		return dummy;
		
	path[n - 4] = 0;
	err = ClassicFSPathMakeFSRef((unsigned char *) path, &ref, nil);
	if (err == noErr)
	{
		err = FSResolveAliasFile(&ref, false, &isFolder, &wasAliased);
		if ((err == noErr) && isFolder)
		{
			err = FSRefMakePath(&ref, (unsigned char *) path, PATH_MAX);
			if (err == noErr)
			{
				strcat(path, MAC_PATH_SEPARATOR);
				strcpy(gPackFolderPath, path);

				return path;
			}
			else
				return dummy;
		}
	}

	char	fold[16];
		
	if (strncmp(Memory.ROMName, "Star Ocean",            10) == 0)
		strcpy(fold, "SOCNSDD1");
	else
	if (strncmp(Memory.ROMName, "STREET FIGHTER ALPHA2", 21) == 0)
	{
		if (Memory.ROMRegion == 1)
			strcpy(fold, "SFA2SDD1");
		else
			strcpy(fold, "SFA2SDD1.E");
	}
	else
	if (strncmp(Memory.ROMName, "STREET FIGHTER ZERO2" , 20) == 0)
		strcpy(fold, "SFZ2SDD1");
	else
	if (strncmp(Memory.ROMName, "SUPER POWER LEAG 4   ", 21) == 0)
		strcpy(fold, "SPL4-SP7");
	else
	if (strncmp(Memory.ROMName, "MOMOTETSU HAPPY      ", 21) == 0)
		strcpy(fold, "SMHT-SP7");
	else
	if (strncmp(Memory.ROMName, "HU TENGAI MAKYO ZERO ", 21) == 0)
		strcpy(fold, "FEOEZSP7");
	else
	if (strncmp(Memory.ROMName, "JUMP TENGAIMAKYO ZERO", 21) == 0)
		strcpy(fold, "SJUMPSP7");
	else
		strcpy(fold, "MISCPACK");
	
	_splitpath(path, drive, dir, fname, ext);
	_makepath(path, drive, dir, fold, "");

	err = ClassicFSPathMakeFSRef((unsigned char *) path, &ref, nil);
	if (err == noErr)
	{
		err = FSResolveAliasFile(&ref, false, &isFolder, &wasAliased);
		if ((err == noErr) && isFolder)
		{
			err = FSRefMakePath(&ref, (unsigned char *) path, PATH_MAX);
			if (err == noErr)
			{
				strcat(path, MAC_PATH_SEPARATOR);
				strcpy(gPackFolderPath, path);

				return path;
			}
			else
				return dummy;
		}
	}
	
	return dummy;
}

const char * S9xGetSnapshotDirectory(void)
{
	//printf("Reading SPC7110 Graphics Pack\n");
	return gPackFolderPath;
}

extern "C" char * osd_GetPackDir(void)
{
	static char	str[PATH_MAX + 1];
	
	strcpy(str, S9xGetPackDirectory());
	
	return str;
}

OSStatus ClassicFSPathMakeFSRef(unsigned char *path, FSRef *ref, Boolean *isdir)
{
	#pragma unused (isdir)
	
	OSStatus	err;
	FSRef		parent;
	char		drive[_MAX_DRIVE], dir[_MAX_DIR], fname[_MAX_FNAME], ext[_MAX_EXT];
	
	_splitpath((char *) path, drive, dir, fname, ext);
	strcat(drive, dir);
	strcat(fname, ext);
	
	err = FSPathMakeRef((unsigned char *) drive, &parent, nil);
	if (err == noErr)
	{
		CFStringRef		str;
		HFSUniStr255	unistr;
	
		str = CFStringCreateWithCString(kCFAllocatorDefault, fname, CFStringGetSystemEncoding());
		if (str)
		{		
			unistr.length = (UniCharCount) CFStringGetLength(str);
			if (unistr.length > 255)
				unistr.length = 255;
			CFStringGetCharacters(str, CFRangeMake(0, unistr.length), unistr.unicode);		
			CFRelease(str);
			
			err = FSMakeFSRefUnicode(&parent, unistr.length, unistr.unicode, CFStringGetSystemEncoding(), ref);
			if ((err != noErr) || (err != fnfErr))
				ref = nil;
		}
		else
			err = 1;
	}
	
	return err;
}

void _splitpath(const char *path, char *drive, char *dir, char *fname, char *ext)
{
	short	x;
	
	drive[0] = '\0';
	fname[0] = '\0';
	ext[0]   = '\0';
	dir[0]   = '\0';
	
	if (strlen(path) < 1)
		return;
	
	x = strlen(path) - 1;
	while (x && (path[x] != MAC_PATH_SEP_CHAR))
		x--;
	
	if (x)
	{
		strcpy(dir, path);
		dir[x + 1] = '\0';
		
		strcpy(fname, path + x + 1);
	}
	else
		strcpy(fname, path);
	
	x = strlen(fname);
	while (x && (fname[x] != '.'))
		x--;
	
	if (x)
	{
		strcpy(ext, fname + x);
		fname[x] = '\0';
	}
}

void _makepath(char *path, const char *drive, const char *dir, const char *fname, const char *ext)
{
	#pragma unused (drive)

	path[0] = '\0';

	if (dir && dir[0])
		strcat(path, dir);
	
	if (fname && fname[0])
		strcat(path, fname);
	
	if (ext && ext[0])
		strcat(path, ext);
}
