/*******************************************************************************
  Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
 
  (c) Copyright 1996 - 2003 Gary Henderson (gary.henderson@ntlworld.com) and
                            Jerremy Koot (jkoot@snes9x.com)

  (c) Copyright 2002 - 2003 Matthew Kendora and
                            Brad Jorsch (anomie@users.sourceforge.net)
 

                      
  C4 x86 assembler and some C emulation code
  (c) Copyright 2000 - 2003 zsKnight (zsknight@zsnes.com),
                            _Demo_ (_demo_@zsnes.com), and
                            Nach (n-a-c-h@users.sourceforge.net)
                                          
  C4 C++ code
  (c) Copyright 2003 Brad Jorsch

  DSP-1 emulator code
  (c) Copyright 1998 - 2003 Ivar (ivar@snes9x.com), _Demo_, Gary Henderson,
                            John Weidman (jweidman@slip.net),
                            neviksti (neviksti@hotmail.com), and
                            Kris Bleakley (stinkfish@bigpond.com)
 
  DSP-2 emulator code
  (c) Copyright 2003 Kris Bleakley, John Weidman, neviksti, Matthew Kendora, and
                     Lord Nightmare (lord_nightmare@users.sourceforge.net

  OBC1 emulator code
  (c) Copyright 2001 - 2003 zsKnight, pagefault (pagefault@zsnes.com)
  Ported from x86 assembler to C by sanmaiwashi

  SPC7110 and RTC C++ emulator code
  (c) Copyright 2002 Matthew Kendora with research by
                     zsKnight, John Weidman, and Dark Force

  S-RTC C emulator code
  (c) Copyright 2001 John Weidman
  
  Super FX x86 assembler emulator code 
  (c) Copyright 1998 - 2003 zsKnight, _Demo_, and pagefault 

  Super FX C emulator code 
  (c) Copyright 1997 - 1999 Ivar and Gary Henderson.



 
  Specific ports contains the works of other authors. See headers in
  individual files.
 
  Snes9x homepage: http://www.snes9x.com
 
  Permission to use, copy, modify and distribute Snes9x in both binary and
  source form, for non-commercial purposes, is hereby granted without fee,
  providing that this license information and copyright notice appear with
  all copies and any derived work.
 
  This software is provided 'as-is', without any express or implied
  warranty. In no event shall the authors be held liable for any damages
  arising from the use of this software.
 
  Snes9x is freeware for PERSONAL USE only. Commercial users should
  seek permission of the copyright holders first. Commercial use includes
  charging money for Snes9x or software derived from Snes9x.
 
  The copyright holders request that bug fixes and improvements to the code
  should be forwarded to them so everyone can benefit from the modifications
  in future versions.
 
  Super NES and Super Nintendo Entertainment System are trademarks of
  Nintendo Co., Limited and its subsidiary companies.
*******************************************************************************/

/*******************************************************************************
  Snes9x custom (Snes9x for Mac OS) by zones
  
  This build was started from John Stiles' port.
*******************************************************************************/

#include "snes9x.h"
#include "memmap.h"
#include "apu.h"
#include "snapshot.h"
#include "soundux.h"

#include "os9-prefix.h"
#include "os9-dialog.h"
#include "os9-file.h"
#include "os9-os.h"
#include "os9-sound.h"
#include "os9-musicbox.h"

uint16	mscbx = ~0;

static void SPCPlayExec(void);
static void SPCPlayFreeze(void);
static void SPCPlayDefrost(void);
static void MusicBoxForceFreeze(void);
static void MusicBoxForceDefrost(void);
static pascal void SoundTask(TMTask *);
static pascal OSStatus MusicBoxEventHandler(EventHandlerCallRef, EventRef, void *);

static volatile Boolean	installed, stopNow;

static int32			mbxVCounter;
static long				delay;

static SAPU				*StoredAPU;
static SAPURegisters	*StoredAPURegisters;
static SSoundData		*StoredSoundData;
static uint8			*StoredIAPURAM;

void MusicBoxDialog(void)
{
	OSStatus	err;
	IBNibRef	nibRef;
	
	if (!cartOpen)
		return;
	
	err = CreateNibReference(CFSTR("SNES9X"), &nibRef);
	if (err == noErr)
	{
		WindowRef	tWindowRef;
		
		err = CreateWindowFromNib(nibRef, CFSTR("MusicBox"), &tWindowRef);
		if (err == noErr)
		{
			EventHandlerRef		eref;
			EventHandlerUPP		eventUPP;
			EventTypeSpec		windowEvents[] = { { kEventClassCommand, kEventCommandProcess } };
			TMTask				tmtask;
			
			if (musicboxmode == kMBXSoundEmulation)
				SPCPlayFreeze();
			else
				MusicBoxForceFreeze();

			for (int i = 0; i < 5; i++)
				controlPad[i] = 0;

			eventUPP = NewEventHandlerUPP(MusicBoxEventHandler);
			err = InstallWindowEventHandler(tWindowRef, eventUPP, 1, windowEvents, (void *) tWindowRef, &eref);
			
			tmtask.tmWakeUp = 0;
			tmtask.tmReserved = 0;
			tmtask.tmAddr = NewTimerUPP(SoundTask);			
			err = InstallXTimeTask((QElemPtr) &tmtask);

			delay = 1000000 / Memory.ROMFramesPerSecond;
			stopNow = false;
			installed = true;

			mbxVCounter = 0;

			err = PrimeTimeTask((QElemPtr) &tmtask, -delay);
			
			MoveWindowPosition(tWindowRef, kWindowMusicBox);
			ShowWindow(tWindowRef);						
			err = RunAppModalLoopForWindow(tWindowRef);						
			HideWindow(tWindowRef);
			SaveWindowPosition(tWindowRef, kWindowMusicBox);
			
			stopNow = true;
			while (installed)
				MPYield();
			
			err = RemoveTimeTask((QElemPtr) &tmtask);
			DisposeTimerUPP(tmtask.tmAddr);
			
			err = RemoveEventHandler(eref);
			DisposeEventHandlerUPP(eventUPP);
			
			ReleaseWindow(tWindowRef);
			
			if (musicboxmode == kMBXSoundEmulation)
				SPCPlayDefrost();
			else
				MusicBoxForceDefrost();
		}
		
		DisposeNibReference(nibRef);
	}
}

static pascal OSStatus MusicBoxEventHandler(EventHandlerCallRef inHandlerCallRef, EventRef inEvent, void *inUserData)
{
    #pragma unused (inHandlerCallRef)
    
    OSStatus	err, result = eventNotHandledErr;
	WindowRef	tWindowRef = (WindowRef) inUserData;

	switch (GetEventClass(inEvent))
	{
		case kEventClassCommand:
			switch (GetEventKind(inEvent))
			{
				case kEventCommandProcess:
					HICommand	cmd;

					err = GetEventParameter(inEvent, kEventParamDirectObject, typeHICommand, nil, sizeof(HICommand), nil, &cmd);
					if (err == noErr)
					{
						switch (cmd.commandID)
						{
							case 'ok  ':
								QuitAppModalLoopForWindow(tWindowRef);
								result = noErr;
						}
					}
			}
	}
	
	return result;
}

static pascal void SoundTask(TMTask *task)
{
	if (!stopNow)
	{
		(musicboxmode == kMBXSoundEmulation) ? SPCPlayExec() : S9xMainLoop();
		PrimeTimeTask((QElemPtr) task, -delay);
	}
	else
		installed = false;
}

static void SPCPlayExec(void)
{
	for (;;)
	{	
		while (APU.Cycles < Settings.HBlankStart)
			APU_EXECUTE1();

		APU.Cycles -= Settings.H_Max;
	
		if (++mbxVCounter >= (Settings.PAL ? SNES_MAX_PAL_VCOUNTER : SNES_MAX_NTSC_VCOUNTER))
		{
			mbxVCounter = 0;
			break;
		}
		
		if (APU.TimerEnabled[2])		// 2: 64kHz
		{
			APU.Timer[2] += 4;
			while (APU.Timer[2] >= APU.TimerTarget[2])
			{
				IAPU.RAM[0xff] = (IAPU.RAM[0xff] + 1) & 0xf;
				APU.Timer[2] -= APU.TimerTarget[2];
				
				#ifdef SPC700_SHUTDOWN		
				IAPU.WaitCounter++;
				IAPU.APUExecuting = true;
				#endif		
			}
		}

		if (mbxVCounter & 1)
		{
			if (APU.TimerEnabled[0])	// 0: 8kHz
			{
				APU.Timer[0]++;
				if (APU.Timer[0] >= APU.TimerTarget[0])
				{
					IAPU.RAM[0xfd] = (IAPU.RAM[0xfd] + 1) & 0xf;
					APU.Timer[0] = 0;
					
					#ifdef SPC700_SHUTDOWN		
					IAPU.WaitCounter++;
					IAPU.APUExecuting = true;
					#endif		    
				}
			}
			
			if (APU.TimerEnabled[1])	// 1: 8kHz
			{
				APU.Timer[1]++;
				if (APU.Timer[1] >= APU.TimerTarget[1])
				{
					IAPU.RAM[0xfe] = (IAPU.RAM[0xfe] + 1) & 0xf;
					APU.Timer[1] = 0;

					#ifdef SPC700_SHUTDOWN		
					IAPU.WaitCounter++;
					IAPU.APUExecuting = true;
					#endif		    
				}
			}
		}
	}
	
	APURegisters.PC = IAPU.PC - IAPU.RAM;
    S9xAPUPackStatus();

	int	soundBuffersToQueue;
	
	soundBuffersToQueue = macBufferCount / 12;
	
	if (soundQueued < soundBuffersToQueue)
	{
		int	sample_count;
		
		sample_count = so.buffer_size;
		if (so.sixteen_bit)
			sample_count >>= 1;

		S9xMixSamples((uint8 *) soundBuffer[soundBufferAt], sample_count);
		if (++soundBufferAt >= macBufferCount)
			soundBufferAt = 0;

		soundQueued++;
	}
}

static void MusicBoxForceFreeze(void)
{
	char	filename[PATH_MAX + 1];
	
	strcpy(filename, S9xGetFreezeFilename(999));
	strcat(filename, ".tmp");
	
	S9xFreezeGame(filename);
}

static void MusicBoxForceDefrost(void)
{
	char	filename[PATH_MAX + 1];
	
	strcpy(filename, S9xGetFreezeFilename(999));
	strcat(filename, ".tmp");
	
	S9xUnfreezeGame(filename);
	remove(filename);
}

static void SPCPlayFreeze(void)
{
	int	i, j;
	
	StoredAPU          = new SAPU;
	StoredAPURegisters = new SAPURegisters;
	StoredSoundData    = new SSoundData;
	StoredIAPURAM      = new uint8 [0x10000];
	
	S9xSetSoundMute(true);
		
	for (i = 0; i < 8; i++)
    {
		SoundData.channels[i].previous16[0] = (int16) SoundData.channels[i].previous[0];
		SoundData.channels[i].previous16[1] = (int16) SoundData.channels[i].previous[1];
    }

	memcpy(StoredAPU,          &APU,          sizeof(SAPU));
	memcpy(StoredAPURegisters, &APURegisters, sizeof(SAPURegisters));
	memcpy(StoredSoundData,    &SoundData,    sizeof(SSoundData));
	memcpy(StoredIAPURAM,      IAPU.RAM,      0x10000);
	
	// Clear needless values : This may be meanless.
	
	StoredSoundData->no_filter = false;
	StoredSoundData->master_volume[0] = StoredSoundData->master_volume[1] = 0;
	StoredSoundData->echo_volume[0] = StoredSoundData->echo_volume[1] = 0;
	StoredSoundData->noise_hertz = 0;
	
	for (i = 0; i < 8; i++)
	{
		StoredSoundData->channels[i].frequency = 0;
		StoredSoundData->channels[i].block = nil;
		StoredSoundData->channels[i].echo_buf_ptr = nil;
		StoredSoundData->channels[i].envxx = 0;
		StoredSoundData->channels[i].next_sample = 0;
		StoredSoundData->channels[i].interpolate = 0;
		StoredSoundData->channels[i].previous[0] = StoredSoundData->channels[i].previous[1] = 0;
		//StoredSoundData->channels[i].last_valid_header = 0;		
		for (j = 0; j < 8; j++)
			StoredSoundData->channels[i].dummy[j] = 0;
	}
	
	S9xSetSoundMute(false);
}

static void SPCPlayDefrost(void)
{
    S9xResetAPU();
    S9xSetSoundMute(true);
	
	memcpy(&APU,          StoredAPU,          sizeof(SAPU));
	memcpy(&APURegisters, StoredAPURegisters, sizeof(SAPURegisters));
	memcpy(&SoundData,    StoredSoundData,    sizeof(SSoundData));
	memcpy(IAPU.RAM,      StoredIAPURAM,      0x10000);
	
	S9xSetSoundMute(false);

	IAPU.PC = IAPU.RAM + APURegisters.PC;
	S9xAPUUnpackStatus();
	
	if (APUCheckDirectPage())
	    IAPU.DirectPage = IAPU.RAM + 0x100;
	else
	    IAPU.DirectPage = IAPU.RAM;

	IAPU.APUExecuting = true;

	S9xFixSoundAfterSnapshotLoad();
	
	delete    StoredAPU;
	delete    StoredAPURegisters;
	delete    StoredSoundData;
	delete [] StoredIAPURAM;
}
