/*******************************************************************************
  Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
 
  (c) Copyright 1996 - 2003 Gary Henderson (gary.henderson@ntlworld.com) and
                            Jerremy Koot (jkoot@snes9x.com)

  (c) Copyright 2002 - 2003 Matthew Kendora and
                            Brad Jorsch (anomie@users.sourceforge.net)
 

                      
  C4 x86 assembler and some C emulation code
  (c) Copyright 2000 - 2003 zsKnight (zsknight@zsnes.com),
                            _Demo_ (_demo_@zsnes.com), and
                            Nach (n-a-c-h@users.sourceforge.net)
                                          
  C4 C++ code
  (c) Copyright 2003 Brad Jorsch

  DSP-1 emulator code
  (c) Copyright 1998 - 2003 Ivar (ivar@snes9x.com), _Demo_, Gary Henderson,
                            John Weidman (jweidman@slip.net),
                            neviksti (neviksti@hotmail.com), and
                            Kris Bleakley (stinkfish@bigpond.com)
 
  DSP-2 emulator code
  (c) Copyright 2003 Kris Bleakley, John Weidman, neviksti, Matthew Kendora, and
                     Lord Nightmare (lord_nightmare@users.sourceforge.net

  OBC1 emulator code
  (c) Copyright 2001 - 2003 zsKnight, pagefault (pagefault@zsnes.com)
  Ported from x86 assembler to C by sanmaiwashi

  SPC7110 and RTC C++ emulator code
  (c) Copyright 2002 Matthew Kendora with research by
                     zsKnight, John Weidman, and Dark Force

  S-RTC C emulator code
  (c) Copyright 2001 John Weidman
  
  Super FX x86 assembler emulator code 
  (c) Copyright 1998 - 2003 zsKnight, _Demo_, and pagefault 

  Super FX C emulator code 
  (c) Copyright 1997 - 1999 Ivar and Gary Henderson.



 
  Specific ports contains the works of other authors. See headers in
  individual files.
 
  Snes9x homepage: http://www.snes9x.com
 
  Permission to use, copy, modify and distribute Snes9x in both binary and
  source form, for non-commercial purposes, is hereby granted without fee,
  providing that this license information and copyright notice appear with
  all copies and any derived work.
 
  This software is provided 'as-is', without any express or implied
  warranty. In no event shall the authors be held liable for any damages
  arising from the use of this software.
 
  Snes9x is freeware for PERSONAL USE only. Commercial users should
  seek permission of the copyright holders first. Commercial use includes
  charging money for Snes9x or software derived from Snes9x.
 
  The copyright holders request that bug fixes and improvements to the code
  should be forwarded to them so everyone can benefit from the modifications
  in future versions.
 
  Super NES and Super Nintendo Entertainment System are trademarks of
  Nintendo Co., Limited and its subsidiary companies.
*******************************************************************************/

/*******************************************************************************
  Snes9x custom (Snes9x for Mac OS) by zones
  
  This build was started from John Stiles' port.
*******************************************************************************/

#include "snes9x.h"
#include "gfx.h"
#include "screenshot.h"

#include <QuickTime.h>
#include "os9-prefix.h"
#include "os9-file.h"
#include "os9-gworld.h"
#include "os9-os.h"
#include "os9-render.h"
#include "os9-screenshot.h"

static void QuickScreenshot(int, int);
static OSErr ExportPicHandleToFile(PicHandle, OSType, OSType, FSSpec *, ScriptCode);

static void QuickScreenshot(int width, int height)
{	
	OSErr		err;
	PicHandle	myPicture;
	FSRef		ref;
	FSSpec		spec;
	FILE		*fs;
	const char	*filename;
	
	myPicture = GetScreenAsPicHandle(width, height, width * ((width <= 256) ? 2 : 1), height * ((height <= 239) ? 2 : 1));
	
	filename = S9xGetPNGFilename();
	fs = fopen(filename, "wb");
	if (fs)
	{
		fclose(fs);
	
		err = ClassicFSPathMakeFSRef((unsigned char *) filename, &ref, nil);
		err = FSGetCatalogInfo(&ref, kFSCatInfoNone, nil, nil, &spec, nil);
		err = ExportPicHandleToFile(myPicture, kQTFileTypePNG, 0, &spec, smSystemScript);
	}

	KillPicture(myPicture);	
}

PicHandle GetScreenAsPicHandle(int width, int height, int destWidth, int destHeight)
{
	PicHandle	myPicture;
	Rect		drawSize, scaleSize;
	GWorldPtr	drawWorld, scaleWorld;
	Byte		*graphicsIn, *graphicsOut; 
	int			row, graphicsRowBytes;

	SetRect(&drawSize,  0, 0, width,     height);
	SetRect(&scaleSize, 0, 0, destWidth, destHeight);
	
	InitGWorld(&drawWorld,  &drawSize,  16);
	InitGWorld(&scaleWorld, &scaleSize, 16);

	graphicsIn  = (Byte *) GFX.Screen;
	graphicsOut = (Byte *) GetPixBaseAddr(GetGWorldPixMap(drawWorld));
	graphicsRowBytes = GetPixRowBytes(GetGWorldPixMap(drawWorld));
	
	for (row = 0; row < height; row++)
	{
		memcpy(graphicsOut, graphicsIn, width * 2);
		
		graphicsIn += 512 * 2;
		
		graphicsOut += graphicsRowBytes;
	}

	if ((scaleSize.right * scaleSize.bottom) < (drawSize.right * drawSize.bottom))
	{
		PrepareForGDrawing(drawWorld);
		CopyBits(GetPortBitMapForCopyBits(drawWorld),  GetPortBitMapForCopyBits(scaleWorld), &drawSize,  &scaleSize, srcCopy | ditherCopy, nil);
		FinishGDrawing(drawWorld);

		PrepareForGDrawing(scaleWorld);
		myPicture = OpenPicture(&scaleSize);
		CopyBits(GetPortBitMapForCopyBits(scaleWorld), GetPortBitMapForCopyBits(scaleWorld), &scaleSize, &scaleSize, srcCopy, nil);
		ClosePicture();
		FinishGDrawing(scaleWorld);
	}
	else
	{
		PrepareForGDrawing(scaleWorld);
		myPicture = OpenPicture(&scaleSize);
		CopyBits(GetPortBitMapForCopyBits(drawWorld),  GetPortBitMapForCopyBits(scaleWorld), &drawSize,  &scaleSize, srcCopy, nil);
		ClosePicture();
		FinishGDrawing(scaleWorld);
	}

	DisposeGWorld(drawWorld);
	DisposeGWorld(scaleWorld);

	return myPicture;
}

void WriteScreenAsPICTResource(FSSpec *file, int destWidth, int destHeight)
{
	short		refNum;
	PicHandle	myPicture;
	
	myPicture = GetScreenAsPicHandle(imageWidth, imageHeight, destWidth, destHeight);
	
	refNum = FSpOpenResFile(file, fsWrPerm);
	AddResource((Handle) myPicture, 'PICT', 128, "\p");
	WriteResource((Handle) myPicture);

	ReleaseResource((Handle) myPicture);
	CloseResFile(refNum);
}

static OSErr ExportPicHandleToFile(PicHandle thePicture, OSType filetype, OSType filecreator, FSSpec *filespec, ScriptCode filescriptcode)
{
    OSErr					err;
	Handle					hdl;
    GraphicsImportComponent gic;
	
	#define	kMaxQuality	1024
    
    hdl = NewHandleClear(512);
	err = MemError();
	if (!err)
	{
		err = HandAndHand((Handle) thePicture, hdl);
		if (!err)
		{
			err = OpenADefaultComponent(GraphicsImporterComponentType, kQTFileTypePicture, &gic);
			if (!err)
			{
				err = GraphicsImportSetQuality(gic, kMaxQuality);
				if (!err)
				{
					err = GraphicsImportSetDataHandle(gic, hdl);
					if (!err)
						err = GraphicsImportExportImageFile(gic, filetype, filecreator, filespec, filescriptcode);
				}
				
				CloseComponent(gic);
			}
		}
		
		DisposeHandle(hdl);
	}
	
    return err;
}

bool8 S9xDoScreenshot(int width, int height)
{
	Settings.TakeScreenshot = false;
	QuickScreenshot(width, height);
	
	return true;
}
